

import typing




class JMLocation(typing.NamedTuple):
	sourceID: typing.Union[str,None]
	lineNo: int
	charPos: int
	jsonPath: str

	@classmethod
	def instantiate(cls, sourceID:typing.Union[str,None], lineNo:int, charPos:int, jsonPath:str):
		if jsonPath.startswith("."):
			jsonPath = jsonPath[1:]
		return cls(sourceID, lineNo, charPos, jsonPath)
	#

	def __str__(self) -> str:
		if self.sourceID:
			return "{}:{}:{}".format(self.sourceID, self.lineNo + 1, self.charPos + 1)
		else:
			return "{}:{}".format(self.lineNo + 1, self.charPos + 1)
	#

#






__all__ = (
	"minValue", "maxValue", "valueIn", "minLength", "maxLength",
	"minSize", "maxSize",
	"matchRegEx", "tcpPort", "absPOSIXPath", "absWindowsPath",
	"ipAddress", "ip4Address", "ip6Address", "hostNameOrIP4Address", "hostNameOrIP6Address", "hostNameOrIPAddress", "hostName",
)



import typing
import re

from .AbstractConstraint import AbstractConstraint





################################################################################################################################

#
# apply to: int, float
#
class _MinValue(AbstractConstraint):

	def __init__(self, value) -> None:
		self.__value = value
	#

	def __call__(self, value) -> typing.Union[str,None]:
		if value >= self.__value:
			return None
		return ">= {}".format(repr(self.__value))
	#

#

# --------------------------------------------------------------------------------------------------------------------------------

#
# apply to: int, float
#
class _MaxValue(AbstractConstraint):

	def __init__(self, value) -> None:
		self.__value = value
	#

	def __call__(self, value) -> typing.Union[str,None]:
		if value <= self.__value:
			return None
		return "<= {}".format(repr(self.__value))
	#

#

################################################################################################################################

#
# apply to: str, int, float, bool
#
class _ValueIn(AbstractConstraint):

	def __init__(self, values) -> None:
		self.__values = values
	#

	def __call__(self, value) -> typing.Union[str,None]:
		if value in self.__values:
			return None
		return "in {}".format(repr(self.__values))
	#

#

################################################################################################################################

#
# apply to: str
#
class _MinLength(AbstractConstraint):

	def __init__(self, value) -> None:
		self.__value = value
	#

	def __call__(self, value) -> typing.Union[str,None]:
		if len(value) >= self.__value:
			return None
		return "length >= {}".format(repr(self.__value))
	#

#

# --------------------------------------------------------------------------------------------------------------------------------

#
# apply to: str
#
class _MaxLength(AbstractConstraint):

	def __init__(self, value) -> None:
		self.__value = value
	#

	def __call__(self, value) -> typing.Union[str,None]:
		if len(value) <= self.__value:
			return None
		return "length <= {}".format(repr(self.__value))
	#

#

# --------------------------------------------------------------------------------------------------------------------------------

#
# apply to: str
#
class _MatchesRegEx(AbstractConstraint):

	def __init__(self, regExPatternStr:str) -> None:
		self.__regExPatternStr = regExPatternStr
		self.__regExPattern = re.compile(regExPatternStr)
	#

	def __call__(self, value) -> typing.Union[str,None]:
		m = self.__regExPattern.match(value)
		if m:
			return None
		return "/{}/".format(repr(self.__regExPatternStr)[1:-1])
	#

#

################################################################################################################################

#
# apply to: JMDict, JMList
#
class _MinSize(AbstractConstraint):

	def __init__(self, value) -> None:
		self.__value = value
	#

	def __call__(self, value) -> typing.Union[str,None]:
		if len(value) >= self.__value:
			return None
		return "size >= {}".format(repr(self.__value))
	#

#

# --------------------------------------------------------------------------------------------------------------------------------

#
# apply to: JMDict, JMList
#
class _MaxSize(AbstractConstraint):

	def __init__(self, value) -> None:
		self.__value = value
	#

	def __call__(self, value) -> typing.Union[str,None]:
		if len(value) <= self.__value:
			return None
		return "size <= {}".format(repr(self.__value))
	#

#

################################################################################################################################

class _ValidTCPPort(AbstractConstraint):

	def __call__(self, value) -> typing.Union[str,None]:
		assert isinstance(value, int)
		if 0 < len(value) <= 65535:
			return None
		return "TCP port"
	#

#

################################################################################################################################

class _AbsolutePOSIXPath(AbstractConstraint):

	def __call__(self, value) -> typing.Union[str,None]:
		assert isinstance(value, str)
		if not value.startswith("/"):
			return "absolute POSIX path"

		if "//" in value:
			return "absolute POSIX path"
		value = value.replace("\\\\", "x")
		for c in value:
			if c in "|\0*?\\<>:":
				return "absolute POSIX path"
		return None
	#

#

# --------------------------------------------------------------------------------------------------------------------------------

class _AbsoluteWindowsPath(AbstractConstraint):

	def __call__(self, value) -> typing.Union[str,None]:
		assert isinstance(value, str)
		m = re.match("[a-zA-Z]:\\", value)
		if m is None:
			return "absolute Windows path"
		if "\\\\" in value:
			return "absolute Windows path"
		value = value.replace("\\\\", "x")
		for c in value:
			if c in "|\0*?/<>:":
				return "absolute Windows path"
		return None
	#

#

################################################################################################################################

class _ValidIP4Address(AbstractConstraint):

	def __call__(self, value) -> typing.Union[str,None]:
		assert isinstance(value, str)
		# https://stackoverflow.com/questions/106179/regular-expression-to-match-dns-hostname-or-ip-address
		m = re.match(r"^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$", value)
		if m:
			return None
		return "TCPv4 address"
	#

#

# --------------------------------------------------------------------------------------------------------------------------------

class _ValidIP6Address(AbstractConstraint):

	def __call__(self, value) -> typing.Union[str,None]:
		assert isinstance(value, str)
		# https://www.oreilly.com/library/view/regular-expressions-cookbook/9781449327453/ch08s17.html
		m = re.match(r"^(?:[A-F0-9]{1,4}:){7}[A-F0-9]{1,4}$", value)
		if m:
			return None
		return "TCPv6 address"
	#

#

################################################################################################################################

class _ValidHostName(AbstractConstraint):

	def __call__(self, value) -> typing.Union[str,None]:
		assert isinstance(value, str)
		# https://stackoverflow.com/questions/106179/regular-expression-to-match-dns-hostname-or-ip-address
		m = re.match(r"^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$", value)
		if m:
			return None
		return "hostname"
	#

#

################################################################################################################################

class _Or(AbstractConstraint):

	def __init__(self, errMsg:str, *constraints:typing.Tuple[AbstractConstraint]) -> None:
		assert isinstance(errMsg, str)
		self.__errMsg = errMsg
		self.__constraints = constraints
	#

	def __call__(self, value) -> typing.Union[str,None]:
		for c in self.__constraints:
			errMsg = c(value)
			if errMsg is None:
				return None
		return self.__errMsg
	#

#

################################################################################################################################







def minSize(value:int) -> AbstractConstraint:
	assert isinstance(value, int)
	return _MinSize(value)
#

def maxSize(value:int) -> AbstractConstraint:
	assert isinstance(value, int)
	return _MaxSize(value)
#

def minValue(value:typing.Union[int,float]) -> AbstractConstraint:
	assert isinstance(value, (int,float))
	return _MinValue(value)
#

def maxValue(value:typing.Union[int,float]) -> AbstractConstraint:
	assert isinstance(value, (int,float))
	return _MaxValue(value)
#

def valueIn(value:typing.Union[set,list,tuple,typing.Any], *args) -> AbstractConstraint:
	if len(args) == 0:
		if str(type(value)) == "<class 'dict_keys'>":
			return _ValueIn(value)
		assert isinstance(value, (set,list,tuple))
		return _ValueIn(value)
	else:
		_values = [ value, *args ]
		return _ValueIn(_values)
#

def minLength(value:int) -> AbstractConstraint:
	assert isinstance(value, int)
	return _MinLength(value)
#

def maxLength(value:int) -> AbstractConstraint:
	assert isinstance(value, int)
	return _MaxLength(value)
#

def matchRegEx(regExPattern:str) -> AbstractConstraint:
	assert isinstance(regExPattern, str)
	return _MatchesRegEx(regExPattern)
#

def tcpPort() -> AbstractConstraint:
	return _ValidTCPPort()
#

def absPOSIXPath() -> AbstractConstraint:
	return _AbsolutePOSIXPath()
#

def absWindowsPath() -> AbstractConstraint:
	return _AbsoluteWindowsPath()
#

def ipAddress() -> AbstractConstraint:
	return _Or("IP address", _ValidIP4Address(), _ValidIP6Address())
#

def ip4Address() -> AbstractConstraint:
	return _ValidIP4Address()
#

def ip6Address() -> AbstractConstraint:
	return _ValidIP6Address()
#

def hostNameOrIP4Address() -> AbstractConstraint:
	return _Or("hostname or IPv4 address", _ValidIP4Address(), _ValidHostName())
#

def hostNameOrIP6Address() -> AbstractConstraint:
	return _Or("hostname or IPv6 address", _ValidIP6Address(), _ValidHostName())
#

def hostNameOrIPAddress() -> AbstractConstraint:
	return _Or("hostname or IP address", _ValidIP4Address(), _ValidIP6Address(), _ValidHostName())
#

def hostName() -> AbstractConstraint:
	return _ValidHostName()
#







import typing
import math

import jk_json



from .JMLocation import JMLocation
from .jclasses import JMDict, JMValue, JMList, _JMProperty







class JsonParserRelaxedModel(jk_json.ParserBase):

	################################################################################################################################
	## Constructor
	################################################################################################################################

	def __init__(self):
		pass
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	# @jk_json.debugDecorator
	def _tryEat_S(self, ctx:jk_json.ParsingContext, ts:jk_json.TokenStream) -> typing.Tuple[bool,typing.Union[JMDict,JMList,None]]:
		# loc = ts.location

		n = ts.skipAll("eol", None)
		hrPath:typing.List[str] = [ "" ]

		bSuccess, parsedData = self._tryEat_JOBJECT(ctx, hrPath, ts)
		if not bSuccess:
			bSuccess, parsedData = self._tryEat_JARRAY(ctx, hrPath, ts)
			if not bSuccess:
				raise jk_json.ParserErrorException(ts.location, "Syntax error: Expecting valid JSON value element")

		n = ts.skipAll("eol", None)

		#if not ts.isEOS:
		#	raise ParserErrorException(ts.location, "Syntax error")

		return (True, parsedData)
	#

	# @jk_json.debugDecorator
	def _tryEat_JANYVALUE(self, ctx:jk_json.ParsingContext, hrPath:typing.List[str], ts:jk_json.TokenStream)  -> typing.Tuple[bool,typing.Union[JMDict,JMList,JMValue,None]]:
		# loc = ts.location

		t1 = ts.peek()
		_loc = JMLocation.instantiate(t1.sourceID, t1.lineNo, t1.charPos, "".join(hrPath))

		if t1.type == "w":
			if t1.text == "null":
				ts.skip()
				return (True, JMValue(_loc, None))
			if t1.text == "true":
				ts.skip()
				return (True, JMValue(_loc, True))
			if t1.text == "false":
				ts.skip()
				return (True, JMValue(_loc, False))
			if t1.text == "Infinity":
				ts.skip()
				return (True, JMValue(_loc, math.inf))
			if t1.text == "NaN":
				ts.skip()
				return (True, JMValue(_loc, math.nan))
			raise jk_json.ParserErrorException(ts.location, "Syntax error: Expecting valid JSON value element")
		if t1.type == "s":
			ts.skip()
			return (True, JMValue(_loc, t1.text))
		if t1.type == "i":
			ts.skip()
			return (True, JMValue(_loc, int(t1.text)))
		if t1.type == "f":
			ts.skip()
			return (True, JMValue(_loc, float(t1.text)))

		bSuccess, parsedData = self._tryEat_JARRAY(ctx, hrPath, ts)
		if bSuccess:
			return (True, parsedData)

		bSuccess, parsedData = self._tryEat_JOBJECT(ctx, hrPath, ts)
		if bSuccess:
			return (True, parsedData)

		if (t1.type == "d") and (t1.text == "-"):
			m = ts.mark()
			ts.skip()
			t2 = ts.peek()
			if (t2.type == "w") and (t2.text == "Infinity"):
				ts.skip()
				return (True, float("-inf"))
			m.resetToMark()

		return (False, None)
	#

	# @jk_json.debugDecorator
	def _tryEat_JARRAY(self, ctx:jk_json.ParsingContext, hrPath:typing.List[str], ts:jk_json.TokenStream) -> typing.Tuple[bool,typing.Union[JMList,None]]:
		# loc = ts.location

		t1 = ts.peek()
		if (t1.type != "d") or (t1.text != "["):
			return (False, None)
		ts.skip()

		bSuccess, _parsedData = self._tryEat_JVALUE_LIST(ctx, hrPath, ts)
		_loc = JMLocation.instantiate(t1.sourceID, t1.lineNo, t1.charPos, "".join(hrPath))
		if bSuccess:
			parsedData = JMList(_loc, _parsedData)
		else:
			parsedData = JMList(_loc, [])

		t1 = ts.peek()
		if (t1.type != "d") or (t1.text != "]"):
			raise jk_json.ParserErrorException(ts.location, "Syntax error: Expecting JSON array elements or ']'")
		ts.skip()

		return (True, parsedData)
	#

	# @jk_json.debugDecorator
	def _tryEat_JVALUE_LIST(self, ctx:jk_json.ParsingContext, hrPath:typing.List[str], ts:jk_json.TokenStream) -> typing.Tuple[bool,typing.Union[list,None]]:
		# loc = ts.location

		retData = []
	
		_index = 0
		hrPath.append("[{}]".format(_index))
		bSuccess, parsedData = self._tryEat_JANYVALUE(ctx, hrPath, ts)
		hrPath.pop()
		if not bSuccess:
			return (False, None)
		retData.append(parsedData)
		_index += 1

		while True:
			t1 = ts.peek()
			if (t1.type != "d") or (t1.text != ","):
				return (True, retData)
			ts.skip()

			hrPath.append("[{}]".format(_index))
			bSuccess, parsedData = self._tryEat_JANYVALUE(ctx, hrPath, ts)
			hrPath.pop()
			if not bSuccess:
				return (True, retData)
			retData.append(parsedData)
			_index += 1
	#

	# @jk_json.debugDecorator
	def _tryEat_JOBJECT(self, ctx:jk_json.ParsingContext, hrPath:typing.List[str], ts:jk_json.TokenStream) -> typing.Tuple[bool,typing.Union[JMDict,None]]:
		# loc = ts.location

		t1 = ts.peek()
		if (t1.type != "d") or (t1.text != "{"):
			return (False, None)
		ts.skip()

		hrPath.append(".")
		bSuccess, _parsedData = self._tryEat_JPROPERTY_LIST(ctx, hrPath, ts)
		hrPath.pop()
		_loc = JMLocation.instantiate(t1.sourceID, t1.lineNo, t1.charPos, "".join(hrPath))
		if bSuccess:
			parsedData = JMDict(_loc, _parsedData)
		else:
			parsedData = JMDict(_loc, {})

		t1 = ts.peek()
		if (t1.type != "d") or (t1.text != "}"):
			raise jk_json.ParserErrorException(ts.location, "Syntax error: Expecting JSON property or '}'")
		ts.skip()

		return (True, parsedData)
	#

	# @jk_json.debugDecorator
	def _tryEat_JPROPERTY_LIST(self, ctx:jk_json.ParsingContext, hrPath:typing.List[str], ts:jk_json.TokenStream) -> typing.Tuple[bool,typing.Union[dict,None]]:
		# loc = ts.location

		retData = {}
		bSuccess, parsedData = self._tryEat_JPROPERTY(ctx, hrPath, ts)
		if not bSuccess:
			return (False, None)

		assert isinstance(parsedData, _JMProperty)
		retData[parsedData._key] = parsedData

		while True:
			t1 = ts.peek()
			if (t1.type != "d") or (t1.text != ","):
				return (True, retData)
			ts.skip()

			bSuccess, parsedData = self._tryEat_JPROPERTY(ctx, hrPath, ts)
			if not bSuccess:
				return (True, retData)

			assert isinstance(parsedData, _JMProperty)
			if parsedData._key in retData:
				if not ctx.allowDuplicatePropertyNames:
					raise jk_json.ParserErrorException(ts.location, "Syntax error: Duplicate property key detected: " + repr(parsedData._key))
			retData[parsedData._key] = parsedData
	#

	# @jk_json.debugDecorator
	def _tryEat_JPROPERTY(self, ctx:jk_json.ParsingContext, hrPath:typing.List[str], ts:jk_json.TokenStream) -> typing.Tuple[bool,typing.Union[_JMProperty,None]]:
		# loc = ts.location

		t1 = ts.peek()
		if t1.type != "s":
			return (False, None)
		ts.skip()

		t2 = ts.peek()
		if (t2.type != "d") or (t2.text != ":"):
			raise jk_json.ParserErrorException(ts.location, "Syntax error: Expecting ':' followed by a property value!")
		ts.skip()

		hrPath.append(t1.text)
		bSuccess, parsedData = self._tryEat_JANYVALUE(ctx, hrPath, ts)
		if bSuccess:
			_loc = JMLocation.instantiate(t1.sourceID, t1.lineNo, t1.charPos, "".join(hrPath))
			hrPath.pop()
			return (True, _JMProperty(_loc, t1.text, parsedData))
		else:
			hrPath.pop()

		raise jk_json.ParserErrorException(ts.location, "Syntax error: Expecting a valid JSON value!")
	#

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	#
	# Parse the tokens. Construct an abstract syntax tree from the tokens provided by the tokenizer.
	#
	# @param	Tokenizer tokenizer		The tokenizer that provides the tokens.
	#
	def parse(self, tokenizer, bDebugging:bool = False, allowDuplicatePropertyNames:bool = False) -> typing.Union[JMDict,JMList,None]:
		if isinstance(tokenizer, jk_json.TokenStream):
			ts = tokenizer
		else:
			ts = jk_json.TokenStream(tokenizer)

		ctx = jk_json.ParsingContext(bDebugging, allowDuplicatePropertyNames)
		bSuccess, data = self._tryEat_S(ctx, ts)
		if bSuccess:
			return data
		else:
			raise Exception()
	#

#











import os
import typing

import jk_typing
import jk_utils
import jk_logging
import jk_json
import jk_prettyprintobj

from .JMLocation import JMLocation
from .AbstractConstraint import AbstractConstraint
from .jclasses import AbstractJMElement, JMDict, JMValue, JMList
from . import constraints








class JMRetriever(object):

	################################################################################################################################
	## Constants
	################################################################################################################################

	################################################################################################################################
	## Constructor
	################################################################################################################################

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	################################################################################################################################
	## Public Static Methods
	################################################################################################################################

	@staticmethod
	def getValueE(jmData:JMDict, *keySequence:str, refPath:str = None) -> JMValue:
		assert isinstance(jmData, JMDict)

		if refPath is not None:
			assert isinstance(refPath, str)
			if keySequence:
				raise Exception("Either specify a key sequence or a reference path!")
			keySequence = refPath.split(".")

		for i in range(0, len(keySequence) - 1):
			key = keySequence[i]
			assert isinstance(key, str)
			if not key or (" " in key):
				if refPath is not None:
					raise Exception("Invalid reference path specified: " + repr(refPath))
				else:
					raise Exception("Invalid key specified: " + repr(key))
			jmData = jmData.getDictE(key)

		key = keySequence[-1]
		if not key or (" " in key):
			if refPath is not None:
				raise Exception("Invalid reference path specified: " + repr(refPath))
			else:
				raise Exception("Invalid key specified: " + repr(key))
		return jmData.getValueE(key)
	#

	@staticmethod
	def getValueN(jmData:JMDict, *keySequence:str, refPath:str = None) -> typing.Union[JMValue,None]:
		assert isinstance(jmData, JMDict)

		if refPath is not None:
			assert isinstance(refPath, str)
			if keySequence:
				raise Exception("Either specify a key sequence or a reference path!")
			keySequence = refPath.split(".")

		for i in range(0, len(keySequence) - 1):
			key = keySequence[i]
			assert isinstance(key, str)
			if not key or (" " in key):
				if refPath is not None:
					raise Exception("Invalid reference path specified: " + repr(refPath))
				else:
					raise Exception("Invalid key specified: " + repr(key))
			jmData = jmData.getDictN(key)
			if jmData is None:
				return None

		key = keySequence[-1]
		if not key or (" " in key):
			if refPath is not None:
				raise Exception("Invalid reference path specified: " + repr(refPath))
			else:
				raise Exception("Invalid key specified: " + repr(key))
		return jmData.getValueN(key)
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	@staticmethod
	def getStrE(jmData:JMDict, *keySequence:str, refPath:str = None) -> str:
		v = JMRetriever.getValueE(jmData, *keySequence, refPath=refPath)
		return v.vStrE()
	#

	@staticmethod
	def getStrN(jmData:JMDict, *keySequence:str, refPath:str = None) -> typing.Union[str,None]:
		v = JMRetriever.getValueN(jmData, *keySequence, refPath=refPath)
		if v is None:
			return None
		return v.vStrN()
	#

	@staticmethod
	def getIntE(jmData:JMDict, *keySequence:str, refPath:str = None) -> int:
		v = JMRetriever.getValueE(jmData, *keySequence, refPath=refPath)
		return v.vIntE()
	#

	@staticmethod
	def getIntN(jmData:JMDict, *keySequence:str, refPath:str = None) -> typing.Union[int,None]:
		v = JMRetriever.getValueN(jmData, *keySequence, refPath=refPath)
		if v is None:
			return None
		return v.vIntN()
	#

	@staticmethod
	def getFloatE(jmData:JMDict, *keySequence:str, refPath:str = None) -> float:
		v = JMRetriever.getValueE(jmData, *keySequence, refPath=refPath)
		return v.vFloatE()
	#

	@staticmethod
	def getFloatN(jmData:JMDict, *keySequence:str, refPath:str = None) -> typing.Union[float,None]:
		v = JMRetriever.getValueN(jmData, *keySequence, refPath=refPath)
		if v is None:
			return None
		return v.vFloatN()
	#

	@staticmethod
	def getBoolE(jmData:JMDict, *keySequence:str, refPath:str = None) -> bool:
		v = JMRetriever.getValueE(jmData, *keySequence, refPath=refPath)
		return v.vBoolE()
	#

	@staticmethod
	def getBoolN(jmData:JMDict, *keySequence:str, refPath:str = None) -> typing.Union[bool,None]:
		v = JMRetriever.getValueN(jmData, *keySequence, refPath=refPath)
		if v is None:
			return None
		return v.vBoolN()
	#

#










__author__ = "Jürgen Knauth"
__version__ = "0.2024.6.29"



import bz2
import gzip
import chardet
import typing

import jk_json



from .JMLocation import JMLocation
from .JsonParserRelaxedModel import JsonParserRelaxedModel as _JsonParserRelaxedModel
from .AbstractConstraint import AbstractConstraint
from .jclasses import AbstractJMElement, JMDict, JMValue, JMList
from . import constraints
from .JMRetriever import JMRetriever





__tokenizerRelaxed = jk_json.TokenizerRelaxed()

__parserRelaxedModel = _JsonParserRelaxedModel()





#
# Deserialize a JSON string: Reconstruct a python data structure from the specified JSON string.
#
# @param	str textToParse		The JSON to parse in binary or string representation. If binary data is specified a simple UTF-8 decoding is performed
#								to get to a string which is then parsed.
# @param	bool bStrict		If ```True``` this parser sticks strictly to the JSON standard. If ```False``` C-style comments
#								are allowed and strings can be specified with single quotes and double quotes.
#								Furthermore NaN, positive and negative infinitiy is supported.
#
def loadModelFromStr(
		textToParse:str,
		bDebugging:bool = False,
		errPrintFunc = None,
		allowDuplicatePropertyNames:bool = False,
		sourceID:str = None,
	) -> typing.Union[JMDict,JMList]:

	assert isinstance(textToParse, str)
	assert isinstance(bDebugging, bool)
	assert isinstance(allowDuplicatePropertyNames, bool)
	if sourceID is not None:
		assert isinstance(sourceID, str)

	# ----

	if isinstance(textToParse, (bytes, bytearray)):
		textToParse = textToParse.decode("utf-8")
	elif not isinstance(textToParse, str):
		raise Exception("Can't decode JSON data from a non-string or non-binary value!")

	tokenizer = __tokenizerRelaxed
	parser = __parserRelaxedModel

	if errPrintFunc and callable(errPrintFunc):
		try:
			ret = parser.parse(tokenizer.tokenize(textToParse, sourceID), bDebugging, allowDuplicatePropertyNames)
		except jk_json.ParserErrorException as ee:
			if sourceID:
				s = sourceID if len(sourceID) < 40 else ("..." + sourceID[-40:])
				prefix = "{}:{} ".format(s, ee.location.lineNo + 1)
			else:
				prefix = "<unknown-source> "
			prefix = "{}:{} ".format(s, ee.location.lineNo + 1)
			errPrintFunc(prefix + ee.textLine.replace("\t", " "))
			errPrintFunc(" " * (len(prefix) + ee.location.charPos + 1) + "ᐃ")
			errPrintFunc(" " * (len(prefix) + ee.location.charPos + 1 - 6) + "╌╌╍╍━━┛")
			raise
	else:
		ret = parser.parse(tokenizer.tokenize(textToParse, sourceID), bDebugging, allowDuplicatePropertyNames)

	assert isinstance(ret, (JMDict,JMList))
	return ret
#

def loadListModelFromStr(
		textToParse:str,
		bDebugging:bool = False,
		errPrintFunc = None,
		allowDuplicatePropertyNames:bool = False,
		sourceID:str = None,
	) -> JMList:

	ret = loadModelFromStr(textToParse, bDebugging, errPrintFunc, allowDuplicatePropertyNames, sourceID)
	assert isinstance(ret, JMList)
	return ret
#

def loadDictModelFromStr(
		textToParse:str,
		bDebugging:bool = False,
		errPrintFunc = None,
		allowDuplicatePropertyNames:bool = False,
		sourceID:str = None,
	) -> JMDict:

	ret = loadModelFromStr(textToParse, bDebugging, errPrintFunc, allowDuplicatePropertyNames, sourceID)
	assert isinstance(ret, JMDict)
	return ret
#



#
# Deserialize a JSON string: Reconstruct a python data structure reading data from the specified JSON file.
#
# @param	str filePath		The path of the file to load.
# @param	bool bStrict		If ```True``` this parser sticks strictly to the JSON standard. If ```False``` C-style comments
#								are allowed and strings can be specified with single quotes and double quotes.
#								Furthermore NaN, positive and negative infinitiy is supported.
#
def loadModelFromFile(
		filePath:str,
		bDebugging:bool = False,
		encoding:str = None,
		autoDetectEncoding:bool = True,
		errPrintFunc = None,
		allowDuplicatePropertyNames:bool = False
	) -> typing.Union[JMDict,JMList]:

	assert isinstance(filePath, str)
	assert isinstance(bDebugging, bool)

	# ----

	if filePath.endswith(".bz2"):
		with bz2.open(filePath, "rb") as f:
			rawData = f.read()
	elif filePath.endswith(".gz"):
		with gzip.open(filePath, "rb") as f:
			rawData = f.read()
	else:
		with open(filePath, "rb") as f:
			rawData = f.read()

	textData = None

	if encoding is None:
		if autoDetectEncoding:
			try:
				if rawData.startswith(b"\xef\xbb\xbf"):		# utf-8 byte order mark
					rawData = rawData[3:]

				textData = rawData.decode("utf-8")

			except:
				encoding = chardet.detect(rawData)["encoding"]
				if encoding is None:
					encoding = "utf-8"

		else:
			encoding = "utf-8"

	if textData is None:
		textData = rawData.decode(encoding)

	return loadDictModelFromStr(textData, bDebugging, errPrintFunc, allowDuplicatePropertyNames, filePath)
#

def loadListModelFromFile(
		filePath:str,
		bDebugging:bool = False,
		encoding:str = None,
		autoDetectEncoding:bool = True,
		errPrintFunc = None,
		allowDuplicatePropertyNames:bool = False
	) -> JMList:

	ret = loadModelFromFile(filePath, bDebugging, encoding, autoDetectEncoding, errPrintFunc, allowDuplicatePropertyNames)
	assert isinstance(ret, JMList)
	return ret
#

def loadDictModelFromFile(
		filePath:str,
		bDebugging:bool = False,
		encoding:str = None,
		autoDetectEncoding:bool = True,
		errPrintFunc = None,
		allowDuplicatePropertyNames:bool = False
	) -> JMDict:

	ret = loadModelFromFile(filePath, bDebugging, encoding, autoDetectEncoding, errPrintFunc, allowDuplicatePropertyNames)
	assert isinstance(ret, JMDict)
	return ret
#



#
# Deserialize a JSON string: Reconstruct a python data structure reading data from the specified BLOB.
#
# @param	bool bStrict		If ```True``` this parser sticks strictly to the JSON standard. If ```False``` C-style comments
#								are allowed and strings can be specified with single quotes and double quotes.
#								Furthermore NaN, positive and negative infinitiy is supported.
#
def loadModelFromBytes(
		rawData:typing.Union[bytearray,bytes],
		bDebugging:bool = False,
		encoding:str = None,
		autoDetectEncoding:bool = True,
		errPrintFunc = None,
		allowDuplicatePropertyNames:bool = False,
		sourceID:str = None,
	) -> typing.Union[JMDict,JMList]:

	assert isinstance(rawData, (bytes,bytearray))
	assert isinstance(bDebugging, bool)

	# ----

	textData = None

	if encoding is None:
		if autoDetectEncoding:
			try:
				if rawData.startswith(b"\xef\xbb\xbf"):		# utf-8 byte order mark
					rawData = rawData[3:]

				textData = rawData.decode("utf-8")

			except:
				encoding = chardet.detect(rawData)["encoding"]
				if encoding is None:
					encoding = "utf-8"

		else:
			encoding = "utf-8"

	if textData is None:
		textData = rawData.decode(encoding)

	return loadDictModelFromStr(textData, bDebugging, errPrintFunc, allowDuplicatePropertyNames, sourceID)
#

def loadListModelFromBytes(
		rawData:typing.Union[bytearray,bytes],
		bDebugging:bool = False,
		encoding:str = None,
		autoDetectEncoding:bool = True,
		errPrintFunc = None,
		allowDuplicatePropertyNames:bool = False,
		sourceID:str = None,
	) -> JMList:

	ret = loadModelFromBytes(rawData, bDebugging, encoding, autoDetectEncoding, errPrintFunc, allowDuplicatePropertyNames, sourceID)
	assert isinstance(ret, JMList)
	return ret
#

def loadDictModelFromBytes(
		rawData:typing.Union[bytearray,bytes],
		bDebugging:bool = False,
		encoding:str = None,
		autoDetectEncoding:bool = True,
		errPrintFunc = None,
		allowDuplicatePropertyNames:bool = False,
		sourceID:str = None,
	) -> JMDict:

	ret = loadModelFromBytes(rawData, bDebugging, encoding, autoDetectEncoding, errPrintFunc, allowDuplicatePropertyNames, sourceID)
	assert isinstance(ret, JMDict)
	return ret
#







import os
import typing

import jk_prettyprintobj



from .JMLocation import JMLocation
from .AbstractConstraint import AbstractConstraint, _packConstraints







AbstractJMElement = typing.NewType("AbstractJElement", object)
JMDict = typing.NewType("JMDict", AbstractJMElement)
JMList = typing.NewType("JMList", AbstractJMElement)
JMValue = typing.NewType("JMValue", AbstractJMElement)







class AbstractJMElement(jk_prettyprintobj.DumpMixin):

	################################################################################################################################
	## Constructor
	################################################################################################################################

	def __init__(self, location:JMLocation) -> None:
		assert isinstance(location, JMLocation)

		self._location = location
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	@property
	def location(self) -> JMLocation:
		return self._location
	#

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	def _dumpVarNames(self) -> typing.List[str]:
		raise NotImplementedError()
	#

	def _buildErrorTypeMismatch(self, typeStr:str) -> Exception:
		assert isinstance(typeStr, str)

		if typeStr[0] in "aeiou":
			typeStr = "an " + typeStr
		else:
			typeStr = "a " + typeStr

		return Exception("Expecting value at {} to be {} ({})".format(
			repr(self._location.jsonPath),
			typeStr,
			str(self._location),
		))
	#

	def _buildConstraintViolation(self, constraintText:str) -> Exception:
		assert isinstance(constraintText, str)

		return Exception("Expecting value at {} to match: {} ({})".format(
			repr(self._location.jsonPath),
			constraintText,
			str(self._location),
		))
	#

	def _buildConstraintViolationIndexed(self, i:int, constraintText:str) -> Exception:
		assert isinstance(i, int)
		assert isinstance(constraintText, str)

		return Exception("Expecting value at {}[{}] to match: {} ({})".format(
			repr(self._location.jsonPath),
			i,
			constraintText,
			str(self._location),
		))
	#

	def _buildErrorTypeMismatchIndexed(self, i:int, typeStr:str) -> Exception:
		assert isinstance(i, int)
		assert isinstance(typeStr, str)

		if typeStr[0] in "aeiou":
			typeStr = "an " + typeStr
		else:
			typeStr = "a " + typeStr

		return Exception("Expecting value at {}[{}] to be {} ({})".format(
			repr(self._location.jsonPath),
			i,
			typeStr,
			str(self._location),
		))
	#

	def _buildErrorMissingKey(self, key:str) -> Exception:
		assert isinstance(key, str)

		return Exception("Expecting key {} at {} ({})".format(
			repr(key),
			self._location.jsonPath,
			str(self._location),
		))
	#

	def fail(self):
		raise Exception("Can't process data at {} ({})".format(
			self._location.jsonPath,
			str(self._location),
		))
	#

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	def ensureIsDictE(self) -> JMDict:
		raise NotImplementedError()
	#

	def ensureIsListE(self) -> JMList:
		raise NotImplementedError()
	#

	def ensureIsValueE(self) -> JMValue:
		raise NotImplementedError()
	#

#





class JMValue(AbstractJMElement, jk_prettyprintobj.DumpMixin):

	################################################################################################################################
	## Constructor
	################################################################################################################################

	#
	# Constructor method.
	#
	def __init__(self,
			location:JMLocation,
			data:typing.Union[int,str,bool,float,None] = None,
		):

		super().__init__(location)

		self._data = data
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	@property
	def data(self):
		return self._data
	#

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	def _dumpVarNames(self) -> typing.List[str]:
		return [
			"location",
			"data",
		]
	#

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	def __str__(self) -> str:
		s = repr(self._data)
		if s.startswith("'") or s.startswith("\""):
			s = s[1:-1]
		return self.__class__.__name__ + "<(" + s + ")>"
	#

	def __repr__(self) -> str:
		s = repr(self._data)
		if s.startswith("'") or s.startswith("\""):
			s = s[1:-1]
		return self.__class__.__name__ + "<(" + s + ")>"
	#

	def ensureIsDictE(self):
		raise self._buildErrorTypeMismatch("object")
	#

	def ensureIsListE(self):
		raise self._buildErrorTypeMismatch("list")
	#

	def ensureIsValueE(self) -> JMValue:
		return self
	#

	# def ensureIsValueStrE(self) -> JMValue:
	# 	if not isinstance(self._data, str):
	# 		raise self._buildErrorTypeMismatch("str")
	# 	return self
	# #

	# def ensureIsValueBoolE(self) -> JMValue:
	# 	if not isinstance(self._data, bool):
	# 		raise self._buildErrorTypeMismatch("bool")
	# 	return self
	# #

	# def ensureIsValueIntE(self) -> JMValue:
	# 	if not isinstance(self._data, int):
	# 		raise self._buildErrorTypeMismatch("int")
	# 	return self
	# #

	# def ensureIsValueIntOrFloatE(self) -> JMValue:
	# 	if not isinstance(self._data, (int,float)):
	# 		raise self._buildErrorTypeMismatch("int,float")
	# 	return self
	# #

	# def ensureIsValueFloatE(self) -> JMValue:
	# 	if not isinstance(self._data, float):
	# 		raise self._buildErrorTypeMismatch("float")
	# 	return self
	# #

	# --------------------------------------------------------------------------------------------------------------------------------

	def vIntN(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> typing.Union[int,None]:
		if self._data is None:
			return None
		if constraints:
			constraints = _packConstraints(constraints)
		if constraints:
			if isinstance(self._data, int):
				errMsg = constraints(self._data)
				if errMsg:
					raise self._buildConstraintViolation(errMsg)
				return self._data
		else:
			if isinstance(self._data, int):
				return self._data
		raise self._buildErrorTypeMismatch("integer")
	#

	def vIntE(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> int:
		if self._data is None:
			raise self._buildErrorTypeMismatch("integer")
		if constraints:
			constraints = _packConstraints(constraints)
		if constraints:
			if isinstance(self._data, int):
				errMsg = constraints(self._data)
				if errMsg:
					raise self._buildConstraintViolation(errMsg)
				return self._data
		else:
			if isinstance(self._data, int):
				return self._data
		raise self._buildErrorTypeMismatch("integer")
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def vFloatN(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> typing.Union[float,None]:
		if self._data is None:
			return None
		if constraints:
			constraints = _packConstraints(constraints)
		if constraints:
			if isinstance(self._data, float):
				errMsg = constraints(self._data)
				if errMsg:
					raise self._buildConstraintViolation(errMsg)
				return self._data
		else:
			if isinstance(self._data, float):
				return self._data
		raise self._buildErrorTypeMismatch("float")
	#

	def vFloatE(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> float:
		if self._data is None:
			raise self._buildErrorTypeMismatch("float")
		if constraints:
			constraints = _packConstraints(constraints)
		if constraints:
			if isinstance(self._data, float):
				errMsg = constraints(self._data)
				if errMsg:
					raise self._buildConstraintViolation(errMsg)
				return self._data
		else:
			if isinstance(self._data, float):
				return self._data
		raise self._buildErrorTypeMismatch("float")
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def vNumericN(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> typing.Union[float,int,None]:
		if self._data is None:
			return None
		if constraints:
			constraints = _packConstraints(constraints)
		if constraints:
			if isinstance(self._data, (int, float)):
				errMsg = constraints(self._data)
				if errMsg:
					raise self._buildConstraintViolation(errMsg)
				return self._data
		else:
			if isinstance(self._data, (int, float)):
				return self._data
		raise self._buildErrorTypeMismatch("numeric")
	#

	def vNumericE(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> typing.Union[float,int]:
		if self._data is None:
			raise self._buildErrorTypeMismatch("numeric")
		if constraints:
			constraints = _packConstraints(constraints)
		if constraints:
			if isinstance(self._data, (int, float)):
				errMsg = constraints(self._data)
				if errMsg:
					raise self._buildConstraintViolation(errMsg)
			return self._data
		else:
			if isinstance(self._data, (int, float)):
				return self._data
		raise self._buildErrorTypeMismatch("numeric")
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def vBoolN(self) -> typing.Union[bool,None]:
		if self._data is None:
			return None
		if isinstance(self._data, bool):
			return self._data
		raise self._buildErrorTypeMismatch("boolean")
	#

	def vBoolE(self) -> bool:
		if self._data is None:
			raise self._buildErrorTypeMismatch("boolean")
		if isinstance(self._data, bool):
			return self._data
		raise self._buildErrorTypeMismatch("boolean")
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def vStrN(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> typing.Union[str,None]:
		if self._data is None:
			return None
		if constraints:
			constraints = _packConstraints(constraints)
		if constraints:
			if isinstance(self._data, str):
				errMsg = constraints(self._data)
				if errMsg:
					raise self._buildConstraintViolation(errMsg)
				return self._data
		else:
			if isinstance(self._data, str):
				return self._data
		raise self._buildErrorTypeMismatch("string")
	#

	def vStrE(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> str:
		if self._data is None:
			raise self._buildErrorTypeMismatch("string")
		if constraints:
			constraints = _packConstraints(constraints)
		if constraints:
			if isinstance(self._data, str):
				errMsg = constraints(self._data)
				if errMsg:
					raise self._buildConstraintViolation(errMsg)
				return self._data
		else:
			if isinstance(self._data, str):
				return self._data
		raise self._buildErrorTypeMismatch("string")
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def vPimitiveN(self,
		) -> typing.Union[str,int,float,bool,None]:
		return self._data
	#

	def vPimitiveE(self,
		) -> typing.Union[str,int,float,bool]:
		if self._data is None:
			raise self._buildErrorTypeMismatch("primitive")
		return self._data
	#

#



#
# This class is used internally only.
#
class _JMProperty(AbstractJMElement, jk_prettyprintobj.DumpMixin):

	################################################################################################################################
	## Constructor
	################################################################################################################################

	#
	# Constructor method.
	#
	def __init__(self,
			location:JMLocation,
			key:str,
			value:AbstractJMElement,
		):

		super().__init__(location)
		assert isinstance(key, str)
		assert isinstance(value, AbstractJMElement)

		self._key = key
		self._data = value
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	@property
	def data(self):
		return self._data
	#

	@property
	def key(self):
		return self._key
	#

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	def _dumpVarNames(self) -> typing.List[str]:
		return [
			"location",
			"key",
			"data",
		]
	#

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	def vDictE(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> JMDict:
		if isinstance(self._data, JMDict):
			self._data._checkConstraints(*constraints)
			return self._data
		raise self._data._buildErrorTypeMismatch("object")
	#

	def vDictN(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> typing.Union[JMDict,None]:
		if isinstance(self._data, JMDict):
			self._data._checkConstraints(*constraints)
			return self._data
		if isinstance(self._data, JMValue):
			if self._data._data is None:
				return None
		raise self._data._buildErrorTypeMismatch("object")
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def vListE(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> typing.Union[JMList,None]:
		if isinstance(self._data, JMList):
			self._data._checkConstraints(*constraints)
			return self._data
		raise self._data._buildErrorTypeMismatch("list")
	#

	def vListN(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> typing.Union[JMList,None]:
		if isinstance(self._data, JMList):
			self._data._checkConstraints(*constraints)
			return self._data
		if isinstance(self._data, JMValue):
			if self._data._data is None:
				return None
		raise self._data._buildErrorTypeMismatch("list")
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def vStrE(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> str:
		if isinstance(self._data, JMValue):
			return self._data.vStrE(*constraints)
		raise self._data._buildErrorTypeMismatch("string")
	#

	def vStrN(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> typing.Union[str,None]:
		if isinstance(self._data, JMValue):
			return self._data.vStrN(*constraints)
		raise self._data._buildErrorTypeMismatch("string")
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def vIntE(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> int:
		if isinstance(self._data, JMValue):
			return self._data.vIntE(*constraints)
		raise self._data._buildErrorTypeMismatch("integer")
	#

	def vIntN(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> typing.Union[int,None]:
		if isinstance(self._data, JMValue):
			return self._data.vIntN(*constraints)
		raise self._data._buildErrorTypeMismatch("integer")
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def vFloatE(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> float:
		if isinstance(self._data, JMValue):
			return self._data.vFloatE(*constraints)
		raise self._data._buildErrorTypeMismatch("float")
	#

	def vFloatN(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> typing.Union[float,None]:
		if isinstance(self._data, JMValue):
			return self._data.vFloatN(*constraints)
		raise self._data._buildErrorTypeMismatch("float")
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def vNumericE(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> typing.Union[float,int]:
		if isinstance(self._data, JMValue):
			return self._data.vNumericE(*constraints)
		raise self._data._buildErrorTypeMismatch("numeric")
	#

	def vNumericN(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> typing.Union[float,int,None]:
		if isinstance(self._data, JMValue):
			return self._data.vNumericN(*constraints)
		raise self._data._buildErrorTypeMismatch("numeric")
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def vBoolE(self) -> bool:
		if isinstance(self._data, JMValue):
			if isinstance(self._data._data, bool):
				return self._data._data
		raise self._data._buildErrorTypeMismatch("boolean")
	#

	def vBoolN(self) -> typing.Union[bool,None]:
		if isinstance(self._data, JMValue):
			if self._data._data is None:
				return None
			if isinstance(self._data._data, bool):
				return self._data._data
		raise self._data._buildErrorTypeMismatch("boolean")
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def vPrimitiveE(self) -> typing.Union[bool,int,float,str]:
		if isinstance(self._data, JMValue):
			if isinstance(self._data._data, (int,float,str,bool)):
				return self._data._data
		raise self._data._buildErrorTypeMismatch("primitive")
	#

	def vPrimitiveN(self) -> typing.Union[bool,int,float,str,None]:
		if isinstance(self._data, JMValue):
			if self._data._data is None:
				return None
			if isinstance(self._data._data, (int,float,str,bool)):
				return self._data._data
		raise self._data._buildErrorTypeMismatch("primitive")
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def vValueE(self) -> JMValue:
		if isinstance(self._data, JMValue):
			return self._data
		raise self._data._buildErrorTypeMismatch("value")
	#

	def vValueN(self) -> typing.Union[JMValue,None]:
		if isinstance(self._data, JMValue):
			if self._data._data is None:
				return None
			return self._data
		raise self._data._buildErrorTypeMismatch("value")
	#

#




class JMList(AbstractJMElement, jk_prettyprintobj.DumpMixin):

	################################################################################################################################
	## Constructor
	################################################################################################################################

	#
	# Constructor method.
	#
	def __init__(self,
			location:JMLocation,
			data:typing.List[AbstractJMElement],
		):

		super().__init__(location)

		self._data:typing.List[AbstractJMElement] = data
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	@property
	def data(self):
		return self._data
	#

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	def _dumpVarNames(self) -> typing.List[str]:
		return [
			"location",
			"data",
		]
	#

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	def __str__(self) -> str:
		s = self._data.__str__()
		return self.__class__.__name__ + "<(" + s[1:-1] + ")>"
	#

	def __repr__(self) -> str:
		s = self._data.__str__()
		return self.__class__.__name__ + "<(" + s[1:-1] + ")>"
	#

	def ensureIsDictE(self) -> JMDict:
		raise self._buildErrorTypeMismatch("object")
	#

	def ensureIsListE(self) -> JMList:
		return self
	#

	def ensureIsValueE(self) -> JMValue:
		raise self._buildErrorTypeMismatch("value")
	#

	def __getitem__(self, index:int) -> typing.Union[None,int,float,bool,str,JMList,JMDict]:
		if not isinstance(index, int):
			raise TypeError("A list index must be of type 'int'!")

		ret = self._data[index]
		if isinstance(ret, JMValue):
			return ret._data
		assert isinstance(ret, (JMList,JMDict))
		return ret
	#

	def __setitem__(self, index, value):
		raise Exception(self.__class__.__name__ + " classes are read only!")
	#

	def __delitem__(self, index):
		raise Exception(self.__class__.__name__ + " classes are read only!")
	#

	def __len__(self) -> int:
		return self._data.__len__()
	#

	def __contains__(self, value) -> bool:
		return value in self._data
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def _checkConstraints(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> None:
		if self._data is None:
			raise self._buildErrorTypeMismatch("list")
		if constraints:
			constraints = _packConstraints(constraints)
		if constraints:
			errMsg = constraints(self._data)
			if errMsg:
				raise self._buildConstraintViolation(errMsg)
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def toStrList(self,
			*constraints:typing.Tuple[AbstractConstraint],
			bAllowNullValues:bool = False,
		) -> typing.List[typing.Union[str,None]]:
		ret = []
		mRetrieveValue = JMValue.vStrN if bAllowNullValues else JMValue.vStrE
		if constraints:
			constraints = _packConstraints(constraints)
			for i, v in enumerate(self._data):
				if isinstance(v, JMValue):
					x = mRetrieveValue(v)
					if x is not None:
						errMsg = constraints(x)
						if errMsg:
							raise v._buildConstraintViolationIndexed(i, errMsg)
					ret.append(x)
				else:
					raise v._buildErrorTypeMismatchIndexed(i, "string")
		else:
			for i, v in enumerate(self._data):
				if isinstance(v, JMValue):
					ret.append(mRetrieveValue(v))
				else:
					raise v._buildErrorTypeMismatchIndexed(i, "string")
		return ret
	#

	def toIntList(self,
			*constraints:typing.Tuple[AbstractConstraint],
			bAllowNullValues:bool = False,
		) -> typing.List[typing.Union[int,None]]:
		ret = []
		mRetrieveValue = JMValue.vIntN if bAllowNullValues else JMValue.vIntE
		if constraints:
			constraints = _packConstraints(constraints)
			for i, v in enumerate(self._data):
				if isinstance(v, JMValue):
					x = mRetrieveValue(v)
					if x is not None:
						errMsg = constraints(x)
						if errMsg:
							raise v._buildConstraintViolationIndexed(i, errMsg)
					ret.append(x)
				else:
					raise v._buildErrorTypeMismatchIndexed(i, "integer")
		else:
			for i, v in enumerate(self._data):
				if isinstance(v, JMValue):
					ret.append(mRetrieveValue(v))
				else:
					raise v._buildErrorTypeMismatchIndexed(i, "integer")
		return ret
	#

	def toFloatList(self,
			*constraints:typing.Tuple[AbstractConstraint],
			bAllowNullValues:bool = False,
		) -> typing.List[typing.Union[float,None]]:
		ret = []
		mRetrieveValue = JMValue.vFloatN if bAllowNullValues else JMValue.vFloatE
		if constraints:
			constraints = _packConstraints(constraints)
			for i, v in enumerate(self._data):
				if isinstance(v, JMValue):
					x = mRetrieveValue(v)
					if x is not None:
						errMsg = constraints(x)
						if errMsg:
							raise v._buildConstraintViolationIndexed(i, errMsg)
					ret.append(x)
				else:
					raise v._buildErrorTypeMismatchIndexed(i, "float")
		else:
			for i, v in enumerate(self._data):
				if isinstance(v, JMValue):
					ret.append(mRetrieveValue(v))
				else:
					raise v._buildErrorTypeMismatchIndexed(i, "float")
		return ret
	#

	def toNumericList(self,
			*constraints:typing.Tuple[AbstractConstraint],
			bAllowNullValues:bool = False,
		) -> typing.List[typing.Union[float,int,None]]:
		ret = []
		mRetrieveValue = JMValue.vNumericN if bAllowNullValues else JMValue.vNumericE
		if constraints:
			constraints = _packConstraints(constraints)
			for i, v in enumerate(self._data):
				if isinstance(v, JMValue):
					x = mRetrieveValue(v)
					if x is not None:
						errMsg = constraints(x)
						if errMsg:
							raise v._buildConstraintViolationIndexed(i, errMsg)
					ret.append(x)
				else:
					raise v._buildErrorTypeMismatchIndexed(i, "numeric")
		else:
			for i, v in enumerate(self._data):
				if isinstance(v, JMValue):
					ret.append(mRetrieveValue(v))
				else:
					raise v._buildErrorTypeMismatchIndexed(i, "numeric")
		return ret
	#

	def toPrimitiveList(self,
			*constraints:typing.Tuple[AbstractConstraint],
			bAllowNullValues:bool = False,
		) -> typing.List[typing.Union[str,float,int,bool,None]]:
		ret = []
		mRetrieveValue = JMValue.vPimitiveN if bAllowNullValues else JMValue.vPimitiveE
		if constraints:
			constraints = _packConstraints(constraints)
			for i, v in enumerate(self._data):
				if isinstance(v, JMValue):
					x = mRetrieveValue(v)
					if x is not None:
						errMsg = constraints(x)
						if errMsg:
							raise v._buildConstraintViolationIndexed(i, errMsg)
					ret.append(x)
				else:
					raise v._buildErrorTypeMismatchIndexed(i, "primitive")
		else:
			for i, v in enumerate(self._data):
				if isinstance(v, JMValue):
					ret.append(mRetrieveValue(v))
				else:
					raise v._buildErrorTypeMismatchIndexed(i, "primitive")
		return ret
	#

	def toDictList(self,
			bAllowNullValues:bool = False,
		) -> typing.List[typing.Union[JMDict,None]]:
		ret = []
		for i, v in enumerate(self._data):
			if bAllowNullValues:
				if v is None:
					ret.append(None)
					continue
			else:
				if isinstance(v, JMDict):
					ret.append(v)
				else:
					raise v._buildErrorTypeMismatchIndexed(i, "object")
		return ret
	#

#



class JMDict(AbstractJMElement, jk_prettyprintobj.DumpMixin):

	################################################################################################################################
	## Constructor
	################################################################################################################################

	#
	# Constructor method.
	#
	def __init__(self,
			location:JMLocation,
			data:typing.Dict[str,_JMProperty],
		):

		super().__init__(location)
		assert isinstance(data, dict)

		self._data:typing.Dict[str,_JMProperty] = data
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	@property
	def data(self):
		return self._data
	#

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	def _dumpVarNames(self) -> typing.List[str]:
		return [
			"location",
			"data",
		]
	#

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	def __str__(self) -> str:
		s = {
			k:v for k,v in self.itemsv()
		}.__str__()
		return self.__class__.__name__ + "<(" + s[1:-1] + ")>"
	#

	def __repr__(self) -> str:
		s = {
			k:v for k,v in self.itemsv()
		}.__str__()
		return self.__class__.__name__ + "<(" + s[1:-1] + ")>"
	#

	def ensureIsDictE(self) -> JMDict:
		return self
	#

	def ensureIsListE(self) -> JMList:
		raise self._buildErrorTypeMismatch("object")
	#

	def ensureIsValueE(self) -> JMValue:
		raise self._buildErrorTypeMismatch("value")
	#

	def __enter__(self) -> JMDict:
		return self
	#

	def __exit__(self, exception_type, exception_value, exception_traceback):
		pass
	#

	def __len__(self) -> int:
		return self._data.__len__()
	#

	def valuesv(self) -> typing.Iterable[typing.Union[int,float,str,bool,None,JMList,JMDict]]:
		for prop in self._data.values():
			assert isinstance(prop, _JMProperty)
			if isinstance(prop._data, JMValue):
				yield prop._data._data
			else:
				yield prop._data
	#

	def values(self) -> typing.Iterable[typing.Union[JMValue,JMList,JMDict]]:
		for prop in self._data.values():
			assert isinstance(prop, _JMProperty)
			yield prop._data
	#

	def keys(self) -> typing.Iterable[str]:
		yield from self._data.keys()
	#

	def itemsv(self) -> typing.Iterable[typing.Tuple[str,typing.Union[int,float,str,bool,None,JMList,JMDict]]]:
		for key, prop in self._data.items():
			assert isinstance(prop, _JMProperty)
			if isinstance(prop._data, JMValue):
				yield key, prop._data._data
			else:
				yield key, prop._data
	#

	def items(self) -> typing.Iterable[typing.Tuple[str,typing.Union[JMValue,JMList,JMDict]]]:
		for key, prop in self._data.items():
			assert isinstance(prop, _JMProperty)
			yield key, prop._data
	#

	def __getitem__(self, key:str) -> typing.Union[JMValue,JMList,JMDict]:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			ret = self._data[key]
			assert isinstance(ret, _JMProperty)
			return ret._data
		raise self._buildErrorMissingKey(key)
	#

	def get(self, key:str) -> typing.Union[JMValue,JMList,JMDict]:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			assert isinstance(prop, _JMProperty)
			return prop._data
		return None
	#

	def getv(self, key:str) -> typing.Union[int,float,str,bool,None,JMList,JMDict]:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			assert isinstance(prop, _JMProperty)
			if isinstance(prop._data, JMValue):
				return prop._data._data
			else:
				return prop._data
		return None
	#

	def __setitem__(self, key, value):
		raise Exception(self.__class__.__name__ + " classes are read only!")
	#

	def __delitem__(self, x):
		raise Exception(self.__class__.__name__ + " classes are read only!")
	#

	def __contains__(self, key:str) -> bool:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		return key in self._data
	#

	def hasNonNull(self, key:str) -> bool:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		v = self._data.get(key)
		if v is None:
			return False

		assert isinstance(v, _JMProperty)

		if v._data is None:
			return False

		if isinstance(v._data, JMValue):
			return v._data._data is not None

		return True
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def _checkConstraints(self,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> None:
		if self._data is None:
			raise self._buildErrorTypeMismatch("object")
		if constraints:
			constraints = _packConstraints(constraints)
		if constraints:
			errMsg = constraints(self._data)
			if errMsg:
				raise self._buildConstraintViolation(errMsg)
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def getDictE(self, key:str,
			*constraints:typing.Tuple[AbstractConstraint],
			defaultValue:typing.Union[JMDict,None] = None,
		) -> JMDict:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			return prop.vDictE(*constraints)
		if defaultValue is not None:
			assert isinstance(defaultValue, JMDict)
			return defaultValue
		raise self._buildErrorMissingKey(key)
	#

	def getDictN(self, key:str,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> typing.Union[JMDict,None]:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			return prop.vDictN(*constraints)
		return None
	#

	################################################################################################################################

	def getListE(self, key:str,
			defaultValue:typing.Union[JMList,None] = None,
		) -> JMList:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			return prop.vListE()
		if defaultValue is not None:
			assert isinstance(defaultValue, JMList)
			return defaultValue
		raise self._buildErrorMissingKey(key)
	#

	def getListN(self, key:str) -> typing.Union[JMList,None]:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			return prop.vListN()
		return None
	#

	################################################################################################################################

	def getStrE(self, key:str,
			*constraints:typing.Tuple[AbstractConstraint],
			defaultValue:typing.Union[str,None] = None,
		) -> str:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			return prop.vStrE(*constraints)
		if defaultValue is not None:
			assert isinstance(defaultValue, str)
			return defaultValue
		raise self._buildErrorMissingKey(key)
	#

	def getStrN(self, key:str,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> typing.Union[str,None]:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			return prop.vStrN(*constraints)			# there is a key but its value might be null
		return None						# no such key
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def getIntE(self, key:str,
			*constraints:typing.Tuple[AbstractConstraint],
			defaultValue:typing.Union[int,None] = None,
		) -> int:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			return prop.vIntE(*constraints)
		if defaultValue is not None:
			assert isinstance(defaultValue, int)
			return defaultValue
		raise self._buildErrorMissingKey(key)
	#

	def getIntN(self, key:str,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> typing.Union[int,None]:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			return prop.vIntN(*constraints)			# there is a key but its value might be null
		return None						# no such key
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def getFloatE(self, key:str,
			*constraints:typing.Tuple[AbstractConstraint],
			defaultValue:typing.Union[int,float,None] = None,
		) -> float:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			return prop.vFloatE(*constraints)
		if defaultValue is not None:
			assert isinstance(defaultValue, (int,float))
			return float(defaultValue)
		raise self._buildErrorMissingKey(key)
	#

	def getFloatN(self, key:str,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> typing.Union[float,None]:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			return prop.vFloatN(*constraints)		# there is a key but its value might be null
		return None						# no such key
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def getBoolE(self, key:str, defaultValue:typing.Union[bool,None] = None) -> bool:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			return prop.vBoolE()
		if defaultValue is not None:
			assert isinstance(defaultValue, bool)
			return defaultValue
		raise self._buildErrorMissingKey(key)
	#

	def getBoolN(self, key:str) -> typing.Union[bool,None]:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			return prop.vBoolN()		# there is a key but its value might be null
		return None						# no such key
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def getPrimitiveE(self, key:str, defaultValue:typing.Union[str,int,float,bool,None] = None) -> typing.Union[str,int,float,bool]:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			return prop.vPrimitiveE()
		if defaultValue is not None:
			assert isinstance(defaultValue, (str,int,float,bool))
			return defaultValue
		raise self._buildErrorMissingKey(key)
	#

	def getPrimitiveN(self, key:str) -> typing.Union[str,int,float,bool,None]:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			return prop.vPrimitiveN()		# there is a key but its value might be null
		return None							# no such key
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def getNumericE(self, key:str,
			*constraints:typing.Tuple[AbstractConstraint],
			defaultValue:typing.Union[int,float,None] = None,
		) -> typing.Union[int,float]:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			return prop.vNumericE(*constraints)
		if defaultValue is not None:
			assert isinstance(defaultValue, (int,float))
			return defaultValue
		raise self._buildErrorMissingKey(key)
	#

	def getNumericN(self, key:str,
			*constraints:typing.Tuple[AbstractConstraint],
		) -> typing.Union[int,float,None]:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			return prop.vNumericN(*constraints)			# there is a key but its value might be null
		return None							# no such key
	#

	# --------------------------------------------------------------------------------------------------------------------------------

	def getValueE(self, key:str) -> JMValue:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			return prop.vValueE()
		raise self._buildErrorMissingKey(key)
	#

	def getValueN(self, key:str) -> typing.Union[JMValue,None]:
		if not isinstance(key, str):
			raise TypeError("A specified key is of type {} but it must be of type 'str'!".format(type(key)))

		if key in self._data:
			prop = self._data[key]
			return prop.vValueE()
		return None
	#

#


